import asyncio
import os

import httpx
import pytest

from granian._granian import BUILD_GIL
from granian.server.embed import Server as EmbeddedGranian


async def app(scope, protocol):
    protocol.response_str(200, [], 'hello')


@pytest.fixture(scope='function')
def loop():
    return asyncio.get_event_loop()


@pytest.fixture(scope='function')
def embed_server(server_port):
    return EmbeddedGranian(app, port=server_port)


@pytest.mark.skipif(not BUILD_GIL, reason='free-threaded Python')
@pytest.mark.skipif(bool(os.environ.get('GITHUB_WORKFLOW')), reason='CI')
def test_embed_server(loop, server_port, embed_server):
    data = {}

    async def client():
        await asyncio.sleep(1.5)

        h = httpx.AsyncClient()
        try:
            data['res'] = await h.get(f'http://localhost:{server_port}')
        finally:
            embed_server.stop()

    server_task = loop.create_task(embed_server.serve())
    loop.run_until_complete(client())

    assert data['res'].status_code == 200
    assert data['res'].text == 'hello'

    loop.run_until_complete(server_task)



import json
import pathlib
import re
from enum import Enum
from typing import Any, Callable, List, Optional, Type, TypeVar, Union

import click

from .constants import HTTPModes, Interfaces, Loops, RuntimeModes, TaskImpl
from .errors import FatalError
from .http import HTTP1Settings, HTTP2Settings
from .log import LogLevels
from .server import Server


_AnyCallable = Callable[..., Any]
FC = TypeVar('FC', bound=Union[_AnyCallable, click.Command])


class Duration(click.IntRange):
    """Custom parameter type for duration strings like '24h', '6m', '2s', '1h30m', etc.

    If the value is a plain number, it will be treated as seconds.
    """

    name = 'duration'
    _multipliers = {'s': 1, 'm': 60, 'h': 60 * 60, 'd': 60 * 60 * 24}
    _pattern = re.compile(r'^(?:(?P<d>\d+)d)?(?:(?P<h>\d+)h)?(?:(?P<m>\d+)m)?(?:(?P<s>\d+)s)?$')

    def __init__(self, min: Optional[int] = None, max: Optional[int] = None) -> None:
        super().__init__(min, max, min_open=False, max_open=False, clamp=False)

    def convert(self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]) -> Any:
        if value is None:
            return value

        if isinstance(value, int):
            seconds = value
        elif isinstance(value, str):
            if value.isnumeric():
                seconds = int(value)
            elif (match := self._pattern.fullmatch(value)) is not None:
                seconds = (
                    int(match.group('d') or 0) * self._multipliers['d']
                    + int(match.group('h') or 0) * self._multipliers['h']
                    + int(match.group('m') or 0) * self._multipliers['m']
                    + int(match.group('s') or 0) * self._multipliers['s']
                )
            else:
                self.fail(f'{value!r} is not a valid duration', param, ctx)
        else:
            self.fail(f'{value!r} is not a valid duration', param, ctx)

        if self.min is not None and seconds < self.min:
            self.fail(f'{value!r} is less than the minimum allowed value of {self.min} seconds', param, ctx)

        if self.max is not None and seconds > self.max:
            self.fail(f'{value!r} is greater than the maximum allowed value of {self.max} seconds', param, ctx)

        return seconds


class EnumType(click.Choice):
    def __init__(self, enum: Enum, case_sensitive=False) -> None:
        self.__enum = enum
        super().__init__(choices=[item.value for item in enum], case_sensitive=case_sensitive)

    def convert(self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]) -> Enum:
        if value is None or isinstance(value, Enum):
            return value

        converted_str = super().convert(value, param, ctx)
        return self.__enum(converted_str)


def _pretty_print_default(value: Optional[bool]) -> Optional[str]:
    if isinstance(value, bool):
        return 'enabled' if value else 'disabled'
    if isinstance(value, Enum):
        return value.value
    return value


def option(*param_decls: str, cls: Optional[Type[click.Option]] = None, **attrs: Any) -> Callable[[FC], FC]:
    attrs['show_envvar'] = True
    if 'default' in attrs:
        attrs['show_default'] = _pretty_print_default(attrs['default'])
    return click.option(*param_decls, cls=cls, **attrs)


@click.command(
    context_settings={'show_default': True},
    help='APP  Application target to serve.  [required]',
)
@click.argument('app', required=True)
@option(
    '--host',
    default='127.0.0.1',
    help='Host address to bind to',
)
@option('--port', type=int, default=8000, help='Port to bind to.')
@option(
    '--uds', type=click.Path(exists=False, writable=True, path_type=pathlib.Path), help='Unix Domain Socket to bind to.'
)
@option(
    '--interface',
    type=EnumType(Interfaces),
    default=Interfaces.RSGI,
    help='Application interface type',
)
@option('--http', type=EnumType(HTTPModes), default=HTTPModes.auto, help='HTTP version')
@option('--ws/--no-ws', 'websockets', default=True, help='Enable websockets handling')
@option('--workers', type=click.IntRange(1), default=1, help='Number of worker processes')
@option(
    '--blocking-threads',
    type=click.IntRange(1),
    help='Number of blocking threads (per worker)',
)
@option(
    '--blocking-threads-idle-timeout',
    type=Duration(10, 600),
    default=30,
    help='The maximum amount of time in seconds (or a human-readable duration) an idle blocking thread will be kept alive',
)
@option('--runtime-threads', type=click.IntRange(1), default=1, help='Number of runtime threads (per worker)')
@option(
    '--runtime-blocking-threads',
    type=click.IntRange(1),
    help='Number of runtime I/O blocking threads (per worker)',
)
@option(
    '--runtime-mode',
    type=EnumType(RuntimeModes),
    default=RuntimeModes.st,
    help='Runtime mode to use (single/multi threaded)',
)
@option('--loop', type=EnumType(Loops), default=Loops.auto, help='Event loop implementation')
@option(
    '--task-impl',
    type=EnumType(TaskImpl),
    default=TaskImpl.asyncio,
    help='Async task implementation to use',
)
@option(
    '--backlog',
    type=click.IntRange(128),
    default=1024,
    help='Maximum number of connections to hold in backlog (globally)',
)
@option(
    '--backpressure',
    type=click.IntRange(1),
    show_default='backlog/workers',
    help='Maximum number of requests to process concurrently (per worker)',
)
@option(
    '--http1-buffer-size',
    type=click.IntRange(8192),
    default=HTTP1Settings.max_buffer_size,
    help='Sets the maximum buffer size for HTTP/1 connections',
)
@option(
    '--http1-header-read-timeout',
    type=click.IntRange(1, 60_000),
    default=HTTP1Settings.header_read_timeout,
    help='Sets a timeout (in milliseconds) to read headers',
)
@option(
    '--http1-keep-alive/--no-http1-keep-alive',
    default=HTTP1Settings.keep_alive,
    help='Enables or disables HTTP/1 keep-alive',
)
@option(
    '--http1-pipeline-flush/--no-http1-pipeline-flush',
    default=HTTP1Settings.pipeline_flush,
    help='Aggregates HTTP/1 flushes to better support pipelined responses (experimental)',
)
@option(
    '--http2-adaptive-window/--no-http2-adaptive-window',
    default=HTTP2Settings.adaptive_window,
    help='Sets whether to use an adaptive flow control for HTTP2',
)
@option(
    '--http2-initial-connection-window-size',
    type=click.IntRange(1024),
    default=HTTP2Settings.initial_connection_window_size,
    help='Sets the max connection-level flow control for HTTP2',
)
@option(
    '--http2-initial-stream-window-size',
    type=click.IntRange(1024),
    default=HTTP2Settings.initial_stream_window_size,
    help='Sets the `SETTINGS_INITIAL_WINDOW_SIZE` option for HTTP2 stream-level flow control',
)
@option(
    '--http2-keep-alive-interval',
    type=click.IntRange(1, 60_000),
    default=HTTP2Settings.keep_alive_interval,
    help='Sets an interval (in milliseconds) for HTTP2 Ping frames should be sent to keep a connection alive',
)
@option(
    '--http2-keep-alive-timeout',
    type=Duration(1),
    default=HTTP2Settings.keep_alive_timeout,
    help='Sets a timeout (in seconds or a human-readable duration) for receiving an acknowledgement of the HTTP2 keep-alive ping',
)
@option(
    '--http2-max-concurrent-streams',
    type=click.IntRange(10),
    default=HTTP2Settings.max_concurrent_streams,
    help='Sets the SETTINGS_MAX_CONCURRENT_STREAMS option for HTTP2 connections',
)
@option(
    '--http2-max-frame-size',
    type=click.IntRange(1024),
    default=HTTP2Settings.max_frame_size,
    help='Sets the maximum frame size to use for HTTP2',
)
@option(
    '--http2-max-headers-size',
    type=click.IntRange(1),
    default=HTTP2Settings.max_headers_size,
    help='Sets the max size of received header frames',
)
@option(
    '--http2-max-send-buffer-size',
    type=click.IntRange(1024),
    default=HTTP2Settings.max_send_buffer_size,
    help='Set the maximum write buffer size for each HTTP/2 stream',
)
@option('--log/--no-log', 'log_enabled', default=True, help='Enable logging')
@option('--log-level', type=EnumType(LogLevels), default=LogLevels.info, help='Log level')
@option(
    '--log-config',
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=pathlib.Path),
    help='Logging configuration file (json)',
)
@option('--access-log/--no-access-log', 'log_access_enabled', default=False, help='Enable access log')
@option('--access-log-fmt', 'log_access_fmt', help='Access log format')
@option(
    '--ssl-certificate',
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=pathlib.Path),
    help='SSL certificate file',
)
@option(
    '--ssl-keyfile',
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=pathlib.Path),
    help='SSL key file',
)
@option('--ssl-keyfile-password', help='SSL key password')
@option(
    '--ssl-ca',
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=pathlib.Path),
    help='Root SSL cerificate file for client verification',
)
@option(
    '--ssl-crl',
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=pathlib.Path),
    help='SSL CRL file(s)',
    multiple=True,
)
@option(
    '--ssl-client-verify/--no-ssl-client-verify',
    default=False,
    help='Verify clients SSL certificates',
)
@option('--url-path-prefix', help='URL path prefix the app is mounted on')
@option(
    '--respawn-failed-workers/--no-respawn-failed-workers',
    default=False,
    help='Enable workers respawn on unexpected exit',
)
@option(
    '--respawn-interval',
    default=3.5,
    help='The number of seconds to sleep between workers respawn',
)
@option(
    '--rss-sample-interval',
    type=Duration(10, 300),
    default=30,
    help='The sample rate in seconds (or a human-readable duration) for the resource monitor',
)
@option(
    '--workers-lifetime',
    type=Duration(60),
    help='The maximum amount of time in seconds (or a human-readable duration) a worker will be kept alive before respawn',
)
@option(
    '--workers-max-rss',
    type=click.IntRange(1),
    help='The maximum amount of memory (in MiB) a worker can consume before respawn',
)
@option(
    '--workers-kill-timeout',
    type=Duration(1, 1800),
    help='The amount of time in seconds (or a human-readable duration) to wait for killing workers that refused to gracefully stop',
    show_default='disabled',
)
@option(
    '--factory/--no-factory',
    default=False,
    help='Treat target as a factory function, that should be invoked to build the actual target',
)
@option(
    '--working-dir',
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=pathlib.Path),
    help='Set the working directory',
)
@option(
    '--env-files',
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=pathlib.Path),
    help='Environment file(s) to load (requires granian[dotenv] extra)',
    multiple=True,
)
@option(
    '--static-path-route',
    default='/static',
    help='Route for static file serving',
)
@option(
    '--static-path-mount',
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=pathlib.Path),
    help='Path to mount for static file serving',
)
@option(
    '--static-path-expires',
    type=Duration(0),
    default=86400,
    help='Cache headers expiration (in seconds or a human-readable duration) for static file serving. 0 to disable.',
)
@option(
    '--reload/--no-reload',
    default=False,
    help="Enable auto reload on application's files changes (requires granian[reload] extra)",
)
@option(
    '--reload-paths',
    type=click.Path(exists=True, file_okay=True, dir_okay=True, readable=True, path_type=pathlib.Path),
    help='Paths to watch for changes',
    show_default='Working directory',
    multiple=True,
)
@option(
    '--reload-ignore-dirs',
    help=(
        'Names of directories to ignore changes for. '
        "Extends the default list of directories to ignore in watchfiles' default filter"
    ),
    multiple=True,
)
@option(
    '--reload-ignore-patterns',
    help=(
        'File/directory name patterns (regex) to ignore changes for. '
        "Extends the default list of patterns to ignore in watchfiles' default filter"
    ),
    multiple=True,
)
@option(
    '--reload-ignore-paths',
    type=click.Path(exists=False, path_type=pathlib.Path),
    help='Absolute paths to ignore changes for',
    multiple=True,
)
@option(
    '--reload-tick',
    type=click.IntRange(50, 5000),
    help='The tick frequency (in milliseconds) the reloader watch for changes',
    default=50,
)
@option(
    '--reload-ignore-worker-failure/--no-reload-ignore-worker-failure',
    default=False,
    help='Ignore worker failures when auto reload is enabled',
)
@option(
    '--process-name',
    help='Set a custom name for processes (requires granian[pname] extra)',
)
@option(
    '--pid-file',
    type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True, path_type=pathlib.Path),
    help='A path to write the PID file to',
)
@click.version_option(message='%(prog)s %(version)s')
def cli(
    app: str,
    host: str,
    port: int,
    uds: Optional[pathlib.Path],
    interface: Interfaces,
    http: HTTPModes,
    websockets: bool,
    workers: int,
    blocking_threads: Optional[int],
    blocking_threads_idle_timeout: int,
    runtime_threads: int,
    runtime_blocking_threads: Optional[int],
    runtime_mode: RuntimeModes,
    loop: Loops,
    task_impl: TaskImpl,
    backlog: int,
    backpressure: Optional[int],
    http1_buffer_size: int,
    http1_header_read_timeout: int,
    http1_keep_alive: bool,
    http1_pipeline_flush: bool,
    http2_adaptive_window: bool,
    http2_initial_connection_window_size: int,
    http2_initial_stream_window_size: int,
    http2_keep_alive_interval: Optional[int],
    http2_keep_alive_timeout: int,
    http2_max_concurrent_streams: int,
    http2_max_frame_size: int,
    http2_max_headers_size: int,
    http2_max_send_buffer_size: int,
    log_enabled: bool,
    log_access_enabled: bool,
    log_access_fmt: Optional[str],
    log_level: LogLevels,
    log_config: Optional[pathlib.Path],
    ssl_certificate: Optional[pathlib.Path],
    ssl_keyfile: Optional[pathlib.Path],
    ssl_keyfile_password: Optional[str],
    ssl_ca: Optional[pathlib.Path],
    ssl_crl: Optional[List[pathlib.Path]],
    ssl_client_verify: bool,
    url_path_prefix: Optional[str],
    respawn_failed_workers: bool,
    respawn_interval: float,
    rss_sample_interval: int,
    workers_lifetime: Optional[int],
    workers_max_rss: Optional[int],
    workers_kill_timeout: Optional[int],
    factory: bool,
    working_dir: Optional[pathlib.Path],
    env_files: Optional[List[pathlib.Path]],
    static_path_route: str,
    static_path_mount: Optional[pathlib.Path],
    static_path_expires: int,
    reload: bool,
    reload_paths: Optional[List[pathlib.Path]],
    reload_ignore_dirs: Optional[List[str]],
    reload_ignore_patterns: Optional[List[str]],
    reload_ignore_paths: Optional[List[pathlib.Path]],
    reload_tick: int,
    reload_ignore_worker_failure: bool,
    process_name: Optional[str],
    pid_file: Optional[pathlib.Path],
) -> None:
    log_dictconfig = None
    if log_config:
        with log_config.open() as log_config_file:
            try:
                log_dictconfig = json.loads(log_config_file.read())
            except Exception:
                print('Unable to parse provided logging config.')
                raise click.exceptions.Exit(1)

    from ._internal import patch_pypath

    patch_pypath(working_dir)

    server = Server(
        app,
        address=host,
        port=port,
        uds=uds,
        interface=interface,
        workers=workers,
        blocking_threads=blocking_threads,
        blocking_threads_idle_timeout=blocking_threads_idle_timeout,
        runtime_threads=runtime_threads,
        runtime_blocking_threads=runtime_blocking_threads,
        runtime_mode=runtime_mode,
        loop=loop,
        task_impl=task_impl,
        http=http,
        websockets=websockets,
        backlog=backlog,
        backpressure=backpressure,
        http1_settings=HTTP1Settings(
            header_read_timeout=http1_header_read_timeout,
            keep_alive=http1_keep_alive,
            max_buffer_size=http1_buffer_size,
            pipeline_flush=http1_pipeline_flush,
        ),
        http2_settings=HTTP2Settings(
            adaptive_window=http2_adaptive_window,
            initial_connection_window_size=http2_initial_connection_window_size,
            initial_stream_window_size=http2_initial_stream_window_size,
            keep_alive_interval=http2_keep_alive_interval,
            keep_alive_timeout=http2_keep_alive_timeout,
            max_concurrent_streams=http2_max_concurrent_streams,
            max_frame_size=http2_max_frame_size,
            max_headers_size=http2_max_headers_size,
            max_send_buffer_size=http2_max_send_buffer_size,
        ),
        log_enabled=log_enabled,
        log_level=log_level,
        log_dictconfig=log_dictconfig,
        log_access=log_access_enabled,
        log_access_format=log_access_fmt,
        ssl_cert=ssl_certificate,
        ssl_key=ssl_keyfile,
        ssl_key_password=ssl_keyfile_password,
        ssl_ca=ssl_ca,
        ssl_crl=ssl_crl,
        ssl_client_verify=ssl_client_verify,
        url_path_prefix=url_path_prefix,
        respawn_failed_workers=respawn_failed_workers,
        respawn_interval=respawn_interval,
        rss_sample_interval=rss_sample_interval,
        workers_lifetime=workers_lifetime,
        workers_max_rss=workers_max_rss,
        workers_kill_timeout=workers_kill_timeout,
        factory=factory,
        working_dir=working_dir,
        env_files=env_files,
        static_path_route=static_path_route,
        static_path_mount=static_path_mount,
        static_path_expires=static_path_expires,
        reload=reload,
        reload_paths=reload_paths,
        reload_ignore_paths=reload_ignore_paths,
        reload_ignore_dirs=reload_ignore_dirs,
        reload_ignore_patterns=reload_ignore_patterns,
        reload_tick=reload_tick,
        reload_ignore_worker_failure=reload_ignore_worker_failure,
        process_name=process_name,
        pid_file=pid_file,
    )

    try:
        server.serve()
    except FatalError:
        raise click.exceptions.Exit(1)


def entrypoint():
    cli(auto_envvar_prefix='GRANIAN')



import ipaddress
from functools import wraps as _wraps


class _Forwarders:
    def __init__(self, trusted_hosts: list[str] | str) -> None:
        self.always_trust: bool = trusted_hosts in ('*', ['*'])
        self.literals: set[str] = set()
        self.hosts: set[ipaddress.IPv4Address | ipaddress.IPv6Address] = set()
        self.networks: set[ipaddress.IPv4Network | ipaddress.IPv6Network] = set()

        if self.always_trust:
            return

        if isinstance(trusted_hosts, str):
            trusted_hosts = _parse_raw_hosts(trusted_hosts)

        for host in trusted_hosts:
            try:
                if '/' in host:
                    self.networks.add(ipaddress.ip_network(host))
                    continue
                self.hosts.add(ipaddress.ip_address(host))
            except ValueError:
                self.literals.add(host)

    def __contains__(self, host: str | None) -> bool:
        if self.always_trust:
            return True

        if not host:
            return False

        try:
            ip = ipaddress.ip_address(host)
            if ip in self.hosts:
                return True
            return any(ip in net for net in self.networks)
        except ValueError:
            return host in self.literals

    def get_client_host(self, x_forwarded_for: str) -> str:
        x_forwarded_for_hosts = _parse_raw_hosts(x_forwarded_for)

        if self.always_trust:
            return x_forwarded_for_hosts[0]

        for host in reversed(x_forwarded_for_hosts):
            if host not in self:
                return host

        return x_forwarded_for_hosts[0]


def _parse_raw_hosts(value: str) -> list[str]:
    return [item.strip() for item in value.split(',')]


def wrap_asgi_with_proxy_headers(app, trusted_hosts: list[str] | str = '127.0.0.1'):
    forwarders = _Forwarders(trusted_hosts)

    @_wraps(app)
    def wrapped(scope, receive, send):
        if scope['type'] == 'lifespan':
            return app(scope, receive, send)

        client_addr = scope.get('client')
        client_host = client_addr[0] if client_addr else None

        if client_host in forwarders:
            headers = dict(scope['headers'])

            if x_forwarded_proto := headers.get(b'x-forwarded-proto', b'').decode('latin1').strip():
                if x_forwarded_proto in {'http', 'https', 'ws', 'wss'}:
                    if scope['type'] == 'websocket':
                        scope['scheme'] = x_forwarded_proto.replace('http', 'ws')
                    else:
                        scope['scheme'] = x_forwarded_proto

            if x_forwarded_for := headers.get(b'x-forwarded-for', b'').decode('latin1'):
                if host := forwarders.get_client_host(x_forwarded_for):
                    scope['client'] = (host, 0)

        return app(scope, receive, send)

    return wrapped


def wrap_wsgi_with_proxy_headers(app, trusted_hosts: list[str] | str = '127.0.0.1'):
    forwarders = _Forwarders(trusted_hosts)

    @_wraps(app)
    def wrapped(scope, resp):
        client_host = scope.get('REMOTE_ADDR')

        if client_host in forwarders:
            if x_forwarded_proto := scope.get('HTTP_X_FORWARDED_PROTO'):
                if x_forwarded_proto in {'http', 'https'}:
                    scope['wsgi.url_scheme'] = x_forwarded_proto

            if x_forwarded_for := scope.get('HTTP_X_FORWARDED_FOR'):
                if host := forwarders.get_client_host(x_forwarded_for):
                    scope['REMOTE_ADDR'] = host

        return app(scope, resp)

    return wrapped



import multiprocessing
import socket
import sys
from functools import wraps
from typing import Any, Callable, Dict, Optional, Tuple

from .._futures import _future_watcher_wrapper, _new_cbscheduler
from .._granian import ASGIWorker, ProcInfoCollector, RSGIWorker, SocketHolder, WSGIWorker
from .._internal import load_env
from .._types import SSLCtx
from ..asgi import LifespanProtocol, _callback_wrapper as _asgi_call_wrap
from ..rsgi import _callback_wrapper as _rsgi_call_wrap, _callbacks_from_target as _rsgi_cbs_from_target
from ..wsgi import _callback_wrapper as _wsgi_call_wrap
from .common import (
    WORKERS_METHODS,
    AbstractServer,
    AbstractWorker,
    HTTP1Settings,
    HTTP2Settings,
    HTTPModes,
    Interfaces,
    RuntimeModes,
    TaskImpl,
    configure_logging,
    logger,
    setproctitle,
)


multiprocessing.allow_connection_pickling()


class WorkerProcess(AbstractWorker):
    _idl = 'PID'

    @staticmethod
    def wrap_target(target):
        @wraps(target)
        def wrapped(
            worker_id,
            process_name,
            callback_loader,
            sock,
            loop_impl,
            log_enabled,
            log_level,
            log_config,
            env_files,
            *args,
            **kwargs,
        ):
            from granian._loops import loops

            if process_name:
                setproctitle.setproctitle(f'{process_name} worker-{worker_id}')

            configure_logging(log_level, log_config, log_enabled)
            load_env(env_files)

            sock, _sso = sock
            if sys.platform == 'win32':
                sock = SocketHolder(_sso.fileno())

            loop = loops.get(loop_impl)
            callback = callback_loader()
            return target(worker_id, callback, sock, loop, *args, **kwargs)

        return wrapped

    def _spawn(self, target, args):
        self.inner = multiprocessing.get_context().Process(name='granian-worker', target=target, args=args)

    def _id(self):
        return self.inner.pid

    def terminate(self):
        self.interrupt_by_parent = True
        self.inner.terminate()

    def kill(self):
        self.interrupt_by_parent = True
        self.inner.kill()


class MPServer(AbstractServer[WorkerProcess]):
    @staticmethod
    @WorkerProcess.wrap_target
    def _spawn_asgi_worker(
        worker_id: int,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: Optional[int],
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: Optional[HTTP1Settings],
        http2_settings: Optional[HTTP2Settings],
        websockets: bool,
        static_path: Optional[Tuple[str, str, Optional[str]]],
        log_access_fmt: Optional[str],
        ssl_ctx: SSLCtx,
        scope_opts: Dict[str, Any],
    ):
        from granian._signals import set_loop_signals

        wcallback = _future_watcher_wrapper(_asgi_call_wrap(callback, scope_opts, {}, log_access_fmt))
        shutdown_event = set_loop_signals(loop)

        worker = ASGIWorker(
            worker_id,
            sock,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)

    @staticmethod
    @WorkerProcess.wrap_target
    def _spawn_asgi_lifespan_worker(
        worker_id: int,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: Optional[int],
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: Optional[HTTP1Settings],
        http2_settings: Optional[HTTP2Settings],
        websockets: bool,
        static_path: Optional[Tuple[str, str, Optional[str]]],
        log_access_fmt: Optional[str],
        ssl_ctx: SSLCtx,
        scope_opts: Dict[str, Any],
    ):
        from granian._signals import set_loop_signals

        lifespan_handler = LifespanProtocol(callback)
        wcallback = _future_watcher_wrapper(
            _asgi_call_wrap(callback, scope_opts, lifespan_handler.state, log_access_fmt)
        )
        shutdown_event = set_loop_signals(loop)

        loop.run_until_complete(lifespan_handler.startup())
        if lifespan_handler.interrupt:
            logger.error('ASGI lifespan startup failed', exc_info=lifespan_handler.exc)
            sys.exit(1)

        worker = ASGIWorker(
            worker_id,
            sock,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)
        loop.run_until_complete(lifespan_handler.shutdown())

    @staticmethod
    @WorkerProcess.wrap_target
    def _spawn_rsgi_worker(
        worker_id: int,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: Optional[int],
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: Optional[HTTP1Settings],
        http2_settings: Optional[HTTP2Settings],
        websockets: bool,
        static_path: Optional[Tuple[str, str, Optional[str]]],
        log_access_fmt: Optional[str],
        ssl_ctx: SSLCtx,
        scope_opts: Dict[str, Any],
    ):
        from granian._signals import set_loop_signals

        callback, callback_init, callback_del = _rsgi_cbs_from_target(callback)
        wcallback = _future_watcher_wrapper(_rsgi_call_wrap(callback, log_access_fmt))
        shutdown_event = set_loop_signals(loop)
        callback_init(loop)

        worker = RSGIWorker(
            worker_id,
            sock,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            websockets,
            static_path,
            *ssl_ctx,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)
        callback_del(loop)

    @staticmethod
    @WorkerProcess.wrap_target
    def _spawn_wsgi_worker(
        worker_id: int,
        callback: Any,
        sock: Any,
        loop: Any,
        runtime_mode: RuntimeModes,
        runtime_threads: int,
        runtime_blocking_threads: Optional[int],
        blocking_threads: int,
        blocking_threads_idle_timeout: int,
        backpressure: int,
        task_impl: TaskImpl,
        http_mode: HTTPModes,
        http1_settings: Optional[HTTP1Settings],
        http2_settings: Optional[HTTP2Settings],
        websockets: bool,
        static_path: Optional[Tuple[str, str, Optional[str]]],
        log_access_fmt: Optional[str],
        ssl_ctx: SSLCtx,
        scope_opts: Dict[str, Any],
    ):
        from granian._signals import set_sync_signals

        wcallback = _wsgi_call_wrap(callback, scope_opts, log_access_fmt)
        shutdown_event = set_sync_signals()

        worker = WSGIWorker(
            worker_id,
            sock,
            runtime_threads,
            runtime_blocking_threads,
            blocking_threads,
            blocking_threads_idle_timeout,
            backpressure,
            http_mode,
            http1_settings,
            http2_settings,
            static_path,
            *ssl_ctx,
        )
        serve = getattr(worker, WORKERS_METHODS[runtime_mode][sock.is_uds()])
        scheduler = _new_cbscheduler(loop, wcallback, impl_asyncio=task_impl == TaskImpl.asyncio)
        serve(scheduler, loop, shutdown_event)

    def _init_shared_socket(self):
        super()._init_shared_socket()
        sock = socket.socket(fileno=self._sfd)
        sock.set_inheritable(True)
        self._sso = sock

    def _write_pidfile(self):
        super()._write_pidfile()
        self._rss_collector = ProcInfoCollector()

    def _unlink_pidfile(self):
        self._sso.detach()
        super()._unlink_pidfile()

    def _handle_rss_signal(self, spawn_target, target_loader):
        wpids = {wrk._id(): wrk for wrk in self.wrks}
        try:
            rss_data = self._rss_collector.memory(list(wpids.keys()))
        except Exception:
            logger.warning('Unable to collect resource usage for workers')
            return
        logger.debug(f'Collected resource usages for workers: {rss_data}')
        to_restart = []
        for wpid, wmem in rss_data.items():
            if wmem >= self.workers_rss:
                wrk = wpids[wpid]
                logger.info(f'worker-{wrk.idx + 1} RSS over threshold, gracefully respawning..')
                to_restart.append(wrk.idx)
        if to_restart:
            self._respawn_workers(to_restart, spawn_target, target_loader, delay=self.respawn_interval)

    def _spawn_worker(self, idx, target, callback_loader) -> WorkerProcess:
        return WorkerProcess(
            parent=self,
            idx=idx,
            target=target,
            args=(
                idx + 1,
                self.process_name,
                callback_loader,
                (self._shd, self._sso),
                self.loop,
                self.log_enabled,
                self.log_level,
                self.log_config,
                self.env_files,
                self.runtime_mode,
                self.runtime_threads,
                self.runtime_blocking_threads,
                self.blocking_threads,
                self.blocking_threads_idle_timeout,
                self.backpressure,
                self.task_impl,
                self.http,
                self.http1_settings,
                self.http2_settings,
                self.websockets,
                self.static_path,
                self.log_access_format if self.log_access else None,
                self.ssl_ctx,
                {'url_path_prefix': self.url_path_prefix},
            ),
        )

    def serve(
        self,
        spawn_target: Optional[Callable[..., None]] = None,
        target_loader: Optional[Callable[..., Callable[..., Any]]] = None,
        wrap_loader: bool = True,
    ):
        if self.interface == Interfaces.WSGI:
            if self.blocking_threads > (multiprocessing.cpu_count() * 2 + 1):
                logger.warning(
                    f'Configuration allows spawning up to {self.blocking_threads} Python threads, '
                    'which seems quite high compared to the number of CPU cores available. '
                    'Consider reviewing your configuration and using `backpressure` to limit '
                    'the concurrency on the Python interpreter. '
                    'If this configuration is intentional, you can safely ignore this message.'
                )

        super().serve(spawn_target, target_loader, wrap_loader)



import copyreg
import sys

from ._granian import ListenerSpec as SocketSpec, SocketHolder


copyreg.pickle(SocketHolder, lambda v: (SocketHolder, v.__getstate__()))
copyreg.pickle(SocketSpec, lambda v: (SocketSpec, v.__getstate__()))

if sys.platform != 'win32':
    from ._granian import UnixListenerSpec as UnixSocketSpec

    copyreg.pickle(UnixSocketSpec, lambda v: (UnixSocketSpec, v.__getstate__()))
else:
    UnixSocketSpec = None



import asyncio
import multiprocessing as mp
import os
import sys
from contextlib import asynccontextmanager
from functools import partial
from pathlib import Path

import httpx
import pytest

from granian import Granian


IS_WIN = sys.platform == 'win32'


def _serve(**kwargs):
    server = Granian(f'tests.apps.{kwargs["interface"]}:app', **kwargs)
    server.serve()


@asynccontextmanager
async def _server(interface, runtime_mode, ws=True, tls=False, task_impl='asyncio', static_mount=False):
    certs_path = Path.cwd() / 'tests' / 'fixtures' / 'tls'
    kwargs = {
        'interface': interface,
        'uds': Path('granian.sock'),
        'loop': 'asyncio',
        'blocking_threads': 1,
        'runtime_mode': runtime_mode,
        'task_impl': task_impl,
        'websockets': ws,
        'workers_kill_timeout': 1,
    }
    if tls:
        if tls == 'private':
            kwargs['ssl_cert'] = certs_path / 'pcert.pem'
            kwargs['ssl_key'] = certs_path / 'pkey.pem'
            kwargs['ssl_key_password'] = 'foobar'  # noqa: S105
        else:
            kwargs['ssl_cert'] = certs_path / 'cert.pem'
            kwargs['ssl_key'] = certs_path / 'key.pem'
    if static_mount:
        kwargs['static_path_mount'] = Path.cwd() / 'tests' / 'fixtures'

    proc = mp.get_context('spawn').Process(target=_serve, kwargs=kwargs)
    proc.start()
    await asyncio.sleep(1.5)

    try:
        yield
    finally:
        proc.terminate()
        proc.join(timeout=2)
        if proc.is_alive():
            proc.kill()


@pytest.fixture(scope='function')
def asgi_server(**extras):
    return partial(_server, 'asgi', **extras)


@pytest.fixture(scope='function')
def rsgi_server(**extras):
    return partial(_server, 'rsgi', **extras)


@pytest.fixture(scope='function')
def wsgi_server(**extras):
    return partial(_server, 'wsgi', **extras)


@pytest.fixture(scope='function')
def server_tls(request):
    return partial(_server, request.param, tls=True)


@pytest.fixture(scope='function')
def http_client():
    transport = httpx.HTTPTransport(uds='granian.sock', verify=False)
    return httpx.Client(transport=transport)


@pytest.mark.asyncio
@pytest.mark.skipif(IS_WIN, reason='no UDS on win')
@pytest.mark.skipif(bool(os.getenv('PGO_RUN')), reason='PGO build')
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_asgi(asgi_server, runtime_mode, http_client):
    async with asgi_server(runtime_mode, ws=False):
        res = http_client.get('http://granian/info?test=true')

    assert res.status_code == 200
    assert res.headers['content-type'] == 'application/json'

    data = res.json()
    assert data['asgi'] == {'version': '3.0', 'spec_version': '2.3'}
    assert data['type'] == 'http'
    assert data['http_version'] == '1.1'
    assert data['scheme'] == 'http'
    assert data['method'] == 'GET'
    assert data['path'] == '/info'
    assert data['query_string'] == 'test=true'
    assert data['headers']['host'] == 'granian'
    assert 'http.response.pathsend' in data['extensions']
    assert data['state']['global'] == 'test'


@pytest.mark.asyncio
@pytest.mark.skipif(IS_WIN, reason='no UDS on win')
@pytest.mark.skipif(bool(os.getenv('PGO_RUN')), reason='PGO build')
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_rsgi(rsgi_server, runtime_mode, http_client):
    async with rsgi_server(runtime_mode, ws=False):
        res = http_client.get('http://granian/info?test=true')

    assert res.status_code == 200
    assert res.headers['content-type'] == 'application/json'

    data = res.json()
    assert data['proto'] == 'http'
    assert data['http_version'] == '1.1'
    assert data['rsgi_version'] == '1.5'
    assert data['scheme'] == 'http'
    assert data['method'] == 'GET'
    assert data['path'] == '/info'
    assert data['query_string'] == 'test=true'
    assert data['headers']['host'] == 'granian'
    assert not data['authority']


@pytest.mark.asyncio
@pytest.mark.skipif(IS_WIN, reason='no UDS on win')
@pytest.mark.skipif(bool(os.getenv('PGO_RUN')), reason='PGO build')
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_wsgi(wsgi_server, runtime_mode, http_client):
    payload = 'body_payload'
    async with wsgi_server(runtime_mode):
        res = http_client.post(
            'http://granian/info?test=true', content=payload, headers=[('test', 'val1'), ('test', 'val2')]
        )

    assert res.status_code == 200
    assert res.headers['content-type'] == 'application/json'

    data = res.json()
    assert data['scheme'] == 'http'
    assert data['method'] == 'POST'
    assert data['path'] == '/info'
    assert data['query_string'] == 'test=true'
    assert data['headers']['HTTP_HOST'] == 'granian'
    assert data['content_length'] == str(len(payload))
    assert data['headers']['HTTP_TEST'] == 'val1,val2'


@pytest.mark.asyncio
@pytest.mark.skipif(IS_WIN, reason='no UDS on win')
@pytest.mark.parametrize('server_tls', ['asgi', 'rsgi', 'wsgi'], indirect=True)
@pytest.mark.parametrize('runtime_mode', ['mt', 'st'])
async def test_https(server_tls, runtime_mode, http_client):
    async with server_tls(runtime_mode, ws=False):
        res = http_client.get('https://granian/info?test=true')

    assert res.status_code == 200
    data = res.json()
    assert data['scheme'] == 'https'
